package com.ys.product.ysmq.front.msg;

import java.net.ConnectException;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;

import com.ys.product.ysmq.front.log.Log;
import com.ys.product.ysmq.front.log.LogFactory;
import com.ys.product.ysmq.front.model.ReadMsgResult;

/**
 * <p>
 * 消费消息的线程,创建-消费-提交在一个线程完成
 * </p>
 */
public class ReadCommitMessageThread extends Thread {
    private static Log log = LogFactory.getLog(ReadCommitMessageThread.class);

    AtomicBoolean isRunning = new AtomicBoolean(true);

    CountDownLatch shutdownLatch = new CountDownLatch(1);

    AbstractConsumerMessage consumerMessage;

    /**
     * 拉取的时间间隔,默认200ms.
     */
    long consumeIntervalTime = 200;

    public ReadCommitMessageThread(AbstractConsumerMessage consumerMessage) {
        super("Consumer Message Thread");
        this.consumerMessage = consumerMessage;
    }

    public ReadCommitMessageThread(AbstractConsumerMessage consumerMessage, long consumeIntervalTime) {
        super("Consumer Message Thread");
        this.consumerMessage = consumerMessage;
        this.consumeIntervalTime = consumeIntervalTime;
    }

    @Override
    public void run() {
        String consumerId = null;
        try {
            // 1.创建消费者,每调用一次返回一个消费者,最多创建10个消费者.
            consumerId = consumerMessage.createConsumer();
        } catch (ConnectException e) {// 网络连接异常
            log.error(String.format("createConsumer Connection timed out,group:%s,consumerId:%s.",
                    consumerMessage.getGroup(), consumerId), e);
        } catch (Exception e) {
            // 异常处理,比如:消息管道内部异常
            log.error(String.format("createConsumer,group:%s,consumerId:%s", consumerMessage.getGroup(), consumerId),
                    e);
        }

        // 循环读取,提交消息
        while (isRunning.get()) {
            try {
                // 2.获取消息,读取-提交模式
                ReadMsgResult readMsgResult = consumerMessage.simpleConsumerMessage();
                // 是正确的读取消息
                if (AbstractConsumerMessage.isCorrectRead(readMsgResult.getCode())) {
                    // 自定义业务处理
                    consumerMessage(readMsgResult.getData());
                } else {// 错误处理
                    readMsgResult = consumerMessage.readMessageErrorHandle(readMsgResult);
                    // 可能会创建新的consumerId
                    consumerId = consumerMessage.getConsumerId();
                    if (AbstractConsumerMessage.isCorrectRead(readMsgResult.getCode())) {
                        // 自定义业务处理
                        consumerMessage(readMsgResult.getData());
                    } // 此时再失败的情况,理论上不会出现,如若出现,暂不处理,不然请求太频繁.执行下一次readCommitMessage
                }
            } catch (ConnectException e) {// 网络连接异常
                log.error(String.format("consumerMessage Connection timed out,group:%s,consumerId:%s",
                        consumerMessage.getGroup(), consumerId), e);
            } catch (Exception e) {
                // 异常处理,比如:消息管道内部异常
                log.error(
                        String.format("consumerMessage,group:%s,consumerId:%s", consumerMessage.getGroup(), consumerId),
                        e);
            } finally {
                // 建议1s左右拉一次,放在这里防止消息管道服务出错的时候,频繁发请求.
                try {
                    Thread.sleep(consumeIntervalTime);
                } catch (InterruptedException e) {
                    log.error(String.format("线程中断,group:%s,consumerId:%s", consumerMessage.getGroup(), consumerId), e);
                }
            }
        }
        shutdownLatch.countDown();
    }

    /**
     * <p>
     * 自定义业务处理
     * </p>
     * 
     * @param data
     *            --获取的消息
     */
    protected void consumerMessage(List<Object> data) {
        consumerMessage.consumerCall(data);
    }

    public void shutdown() {
        try {
            isRunning.set(false);
            this.interrupt();
            shutdownLatch.await();
        } catch (InterruptedException e) {
            throw new Error("Interrupted when shutting down consumer worker thread.");
        }
    }
}
